#include "runtime/runtime.h"
#include "runtime/gfx.h"

#include "runtime/renderer_api.h"

typedef struct {

    unsigned int prog;
} vy_gl_pipeline;

#define NUM_SLOTS 256

typedef struct {
    uint32_t generation_in_use[NUM_SLOTS];
    vy_gl_pipeline pipelines[NUM_SLOTS];
} vy_pipeline_storage;

static vy_pipeline_storage _storage;

static vy_gfx_pipeline_id StorePipeline(vy_gl_pipeline pipeline) {
    /* Search for free slot */
    uint32_t slot = NUM_SLOTS;
    for (uint32_t i = 0; i < NUM_SLOTS; ++i) {
        if ((_storage.generation_in_use[i] & 0x1) == 0) {
            slot = i;
            break;
        }
    }
    if (slot == NUM_SLOTS) {
        vyReportError("GL_GFX", "Ran out of pipeline storage slots");
        return (vy_gfx_pipeline_id){0};
    }

    uint32_t generation = _storage.generation_in_use[slot] >> 1;
    generation          = (generation + 1) & 0x1;

    _storage.pipelines[slot]         = pipeline;
    _storage.generation_in_use[slot] = (generation << 1) | 0x1;

    vy_gfx_pipeline_id id;
    id.index = (generation << 27) | slot;
    return id;
}

static void ReleasePipelineSlot(vy_gfx_pipeline_id id) {
    uint32_t slot = id.index & 0x08ffffff;
    uint32_t gen  = (id.index >> 27) & 0x1f;
    if (slot >= NUM_SLOTS)
        return;
    gen = gen << 1 | 0x1;
    if (_storage.generation_in_use[slot] == gen)
        _storage.generation_in_use[slot] &= ~0x1;
}

VY_DLLEXPORT vy_gfx_pipeline_id
vyCompileComputePipeline(const vy_compute_pipeline_info *info) {
#if 0
    char info_log[512];

    GLuint prog   = glCreateProgram();
    GLuint shader = glCreateShader(GL_COMPUTE_SHADER);

    GLchar *code   = (GLchar *)info->compute_source;
    GLint code_len = (GLint)info->compute_source_length;
    glShaderSource(shader, 1, (const GLchar **)&code, &code_len);
    glCompileShader(shader);

    GLint success;
    glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
    if (!success) {
        glGetShaderInfoLog(shader, 512, NULL, info_log);
        vyReportError("GL_GFX",
                      "Failed to compile compute shader\n%s",
                      info_log);
        glDeleteProgram(prog);
        glDeleteShader(shader);
        return (vy_gfx_pipeline_id){0};
    }
    glAttachShader(prog, shader);
    glLinkProgram(prog);
    glGetProgramiv(prog, GL_LINK_STATUS, &success);
    if (!success) {
        glGetProgramInfoLog(prog, 512, NULL, info_log);
        vyReportError("GL_GFX", "Failed to link compute shader\n%s", info_log);
        glDeleteShader(shader);
        glDeleteProgram(prog);
        return (vy_gfx_pipeline_id){0};
    }
    glDeleteShader(shader);
    #endif

    vy_gl_pipeline pipeline;
    pipeline.prog = 0;
    return StorePipeline(pipeline);
}

VY_DLLEXPORT vy_gfx_pipeline_id
vyCompileGraphicsPipeline(const vy_graphics_pipeline_info *info) {

    #if 0
    char info_log[512];
    GLuint prog            = glCreateProgram();
    GLuint vertex_shader   = glCreateShader(GL_VERTEX_SHADER);
    GLuint fragment_shader = glCreateShader(GL_FRAGMENT_SHADER);

    GLchar *code   = (GLchar *)info->vertex_source;
    GLint code_len = (GLint)info->vertex_source_length;
    glShaderSource(vertex_shader, 1, (const GLchar **)&code, &code_len);
    glCompileShader(vertex_shader);

    GLint success;
    glGetShaderiv(vertex_shader, GL_COMPILE_STATUS, &success);
    if (!success) {
        glGetShaderInfoLog(vertex_shader, 512, NULL, info_log);
        vyReportError("GL_GFX",
                      "Failed to compile vertex shader\n%s",
                      info_log);
        glDeleteShader(vertex_shader);
        glDeleteShader(fragment_shader);
        glDeleteProgram(prog);
        return (vy_gfx_pipeline_id){0};
    }
    glAttachShader(prog, vertex_shader);

    code     = (GLchar *)info->fragment_source;
    code_len = (GLint)info->fragment_source_length;
    glShaderSource(fragment_shader, 1, (const GLchar **)&code, &code_len);
    glCompileShader(fragment_shader);

    glGetShaderiv(fragment_shader, GL_COMPILE_STATUS, &success);
    if (!success) {
        glGetShaderInfoLog(fragment_shader, 512, NULL, info_log);
        vyReportError("GFX", "Failed to compile fragment shader\n%s", info_log);
        glDeleteShader(fragment_shader);
        glDeleteShader(fragment_shader);
        glDeleteProgram(prog);
        return (vy_gfx_pipeline_id){0};
    }
    glAttachShader(prog, fragment_shader);

    glLinkProgram(prog);
    glGetProgramiv(prog, GL_LINK_STATUS, &success);
    if (!success) {
        glGetProgramInfoLog(prog, 512, NULL, info_log);
        vyReportError("GFX", "Failed to link graphics shader\n%s", info_log);
        glDeleteShader(vertex_shader);
        glDeleteShader(fragment_shader);
        glDeleteProgram(prog);
        return (vy_gfx_pipeline_id){0};
    }

    glDeleteShader(vertex_shader);
    glDeleteShader(fragment_shader);
    #endif

    vy_gl_pipeline pipeline;
    pipeline.prog = 0;
    return StorePipeline(pipeline);
}
